/*
 * fermiburn, configurable PMDA
 *
 * Written by Troy Dawson on January 2, 2001
 *
 */

#ident "$Id: fermiburn.c,v 0.5 2001/01/04 10:00:00 dawson Exp $"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <pcp/pmapi.h>
#include <pcp/pmda.h>
#include "domain.h"
#include "fermiburn.h"

static char buf[4096];

/*
 * fermiburn PMDA
 *
 */

/*
 * all metrics supported in this PMDA - one table entry for each
 */

static pmdaMetric metrictab[] = {
/* bonnie stage1write */
    { NULL, 
      { PMDA_PMID(1,0), PM_TYPE_32, PM_INDOM_NULL, PM_SEM_INSTANT, 
        PMDA_PMUNITS(0, 0, 0, 0, 0, 0) }, },
/* bonnie stage1read */
    { NULL, 
      { PMDA_PMID(1,1), PM_TYPE_32, PM_INDOM_NULL, PM_SEM_INSTANT, 
        PMDA_PMUNITS(0, 0, 0, 0, 0, 0) }, },
/* bonnie stage2write */
    { NULL, 
      { PMDA_PMID(1,2), PM_TYPE_32, PM_INDOM_NULL, PM_SEM_INSTANT, 
        PMDA_PMUNITS(0, 0, 0, 0, 0, 0) }, },
/* bonnie stage2read */
    { NULL, 
      { PMDA_PMID(1,3), PM_TYPE_32, PM_INDOM_NULL, PM_SEM_INSTANT, 
        PMDA_PMUNITS(0, 0, 0, 0, 0, 0) }, },
/* seti stage1 */
    { NULL, 
      { PMDA_PMID(2,0), PM_TYPE_FLOAT, PM_INDOM_NULL, PM_SEM_INSTANT, 
        PMDA_PMUNITS(0, 0, 0, 0, 0, 0) }, },
/* seti stage2 */
    { NULL, 
      { PMDA_PMID(2,1), PM_TYPE_FLOAT, PM_INDOM_NULL, PM_SEM_INSTANT, 
        PMDA_PMUNITS(0, 0, 0, 0, 0, 0) }, },
};

seti get_seti()
{
	int i;
	int n;
	int nbufindex;
	float f;
	char *bufindex[64];
	char *temp;
	char *temptoken;
	seti newseti={11.111,22.222};
	
	n = get_file("seti/stage1/", "state.sah");
	buf[sizeof(buf)-1] = '\0';
	n = 1000;
	nbufindex = 0;
	bufindex[nbufindex++] = &buf[0];
	for (i=0; i < n; i++) {
		if (buf[i] == '\n') {
			buf[i] = '\0';
			bufindex[nbufindex++] = buf + i + 1;
		}
	}

	for ( i=0; i < nbufindex ; i++ ) {
		temp="";
		if (strncmp("prog", bufindex[i], 4) == 0 ) {
			temp = strtok(bufindex[i]," ");
			temptoken = strtok(temp,"=");
			temptoken = strtok(NULL," ");
			f = atof(temptoken);
			newseti.stage1=100*f;	
			i = nbufindex;
		}
	}	

	n = get_file("seti/stage2/", "state.sah");	
	buf[sizeof(buf)-1] = '\0';
	n = 1000;
	nbufindex = 0;
	bufindex[nbufindex++] = &buf[0];
	for (i=0; i < n; i++) {
		if (buf[i] == '\n') {
			buf[i] = '\0';
			bufindex[nbufindex++] = buf + i + 1;
		}
	}

	for ( i=0; i < nbufindex ; i++ ) {
		temp="";
		if (strncmp("prog", bufindex[i], 4) == 0 ) {
			temp = strtok(bufindex[i]," ");
			temptoken = strtok(temp,"=");
			temptoken = strtok(NULL," ");
			f = atof(temptoken);
			newseti.stage2=100*f;	
			i = nbufindex;
		}
	}	

	return newseti;
}

bonnie get_bonnie()
{
	int i;
	int n;
	int nbufindex;
	char *bufindex[64];
	char *temp;
	char *temptoken;
	bonnie newbonnie={1111,2222,3333,4444};
	
	n = get_file("root/test/bonnie_current.", "stage1");
	
	buf[sizeof(buf)-1] = '\0';

	nbufindex = 0;
	bufindex[nbufindex++] = &buf[0];
	for (i=0; i < n; i++) {
		if (buf[i] == '\n') {
			buf[i] = '\0';
			bufindex[nbufindex++] = buf + i + 1;
		}
	}
	
	for ( i=0; i < nbufindex ; i++ ) {
		temp="";
		if (strncmp("wKb/sec", bufindex[i], 7) == 0 ) {
			temp = strtok(bufindex[i+1]," ");
			temptoken = strtok(temp,"\t");
			newbonnie.stage1write = atoi(temptoken);
			temptoken = strtok(NULL," ");
			newbonnie.stage1read = atoi(temptoken);
		}
	}	

	n = get_file("root/test/bonnie_current.", "stage2");
	
	buf[sizeof(buf)-1] = '\0';

	nbufindex = 0;
	bufindex[nbufindex++] = &buf[0];
	for (i=0; i < n; i++) {
		if (buf[i] == '\n') {
			buf[i] = '\0';
			bufindex[nbufindex++] = buf + i + 1;
		}
	}
	
	for ( i=0; i < nbufindex ; i++ ) {
		temp="";
		if (strncmp("wKb/sec", bufindex[i], 7) == 0 ) {
			temp = strtok(bufindex[i+1]," ");
			temptoken = strtok(temp,"\t");
			newbonnie.stage2write = atoi(temptoken);
			temptoken = strtok(NULL," ");
			newbonnie.stage2read = atoi(temptoken);
		}
	}	

	return newbonnie;
}


/*
 * Get the contents of a file and return them
 */
int get_file(char *middle, char *end){

	int fd;
	int n;
	int i;
	char s[1024]="/";
	
/*
 * create the new string, the end result being the actual file name
 */
	strcat(s,middle);
	strcat(s,end);


/*
 * read in the file into the buffer buf
 */
	if ((fd = open(s, O_RDONLY)) < 0) {
		return -1;
	}
	n = read(fd, buf, sizeof(buf));
	close(fd);
	return n;
}


/*
 * callback provided to pmdaFetch
 */
static int
fermiburn_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
{
   seti seti1;
   bonnie bonnie1;
    __pmID_int		*idp = (__pmID_int *)&(mdesc->m_desc.pmid);

    if ((idp->cluster < 1 || idp->cluster > 2) || (idp->item < 0 || idp->item > 3))
	return PM_ERR_PMID;
    else if (inst != PM_IN_NULL)
	return PM_ERR_INST;

	if (idp->cluster == 1) {	/*fermiburn.lm75*/
	    bonnie1=get_bonnie();
	    switch (idp->item) {
		case 0:
			atom->l = bonnie1.stage1write;
			break ;
		case 1:
			atom->l = bonnie1.stage1read;
			break ;
		case 2:
			atom->l = bonnie1.stage2write;
			break ;
		case 3:
			atom->l = bonnie1.stage2read;
			break ;
		default:
			printf ("The magic 8 ball says: Not likely \n");
			break ;
		}
	}
	if (idp->cluster == 2) {	/*fermiburn.lm79*/
	    seti1=get_seti();
	    switch (idp->item) {
		case 0:
			atom->f = seti1.stage1;
			break ;
		case 1:
			atom->f = seti1.stage2;
			break ;
		default:
			printf ("Zis one ... e be tricky\n");
			break ;
		}
	}

    return 0;
}

/*
 * Initialise the agent (both daemon and DSO).
 */
void 
fermiburn_init(pmdaInterface *dp)
{
    pmdaSetFetchCallBack(dp, fermiburn_fetchCallBack);

    pmdaInit(dp, NULL, 0, 
	     metrictab, sizeof(metrictab)/sizeof(metrictab[0]));
}

static void
usage(void)
{
    fprintf(stderr, "Usage: %s [options]\n\n", pmProgname);
    fputs("Options:\n"
	  "  -d domain    use domain (numeric) for metrics domain of PMDA\n"
	  "  -l logfile   write log into logfile rather than using default log name\n",
	      stderr);		
    exit(1);
}

/*
 * Set up the agent if running as a daemon.
 */
int
main(int argc, char **argv)
{
    int			err = 0;
    int			c = 0;
    pmdaInterface	desc;
    char		*p;

    /* trim cmd name of leading directory components */
    pmProgname = argv[0];
    for (p = pmProgname; *p; p++) {
	if (*p == '/')
	    pmProgname = p+1;
    }

    pmdaDaemon(&desc, PMDA_INTERFACE_2, pmProgname, FERMIBURN,
		"fermiburn.log", "/var/pcp/pmdas/fermiburn/help");

    if ((c = pmdaGetOpt(argc, argv, "D:d:l:?", &desc, &err)) != EOF)
    	err++;
   
    if (err)
    	usage();

    pmdaOpenLog(&desc);
    fermiburn_init(&desc);
    pmdaConnect(&desc);
    pmdaMain(&desc);

    exit(0);
    /*NOTREACHED*/
}
